# import json
# import re
# from typing import Type
#
# from langchain_community.chat_models import ChatVertexAI
# from langchain_core.exceptions import OutputParserException
# from langchain_core.output_parsers import BaseOutputParser, StrOutputParser
# from langchain_core.prompts import ChatPromptTemplate
# from pydantic import ValidationError
#
# from voice_stream.types import T
#
#
# def reduce_schema(schema: dict) -> dict:
#     if "title" in schema:
#         del schema["title"]
#     if "type" in schema:
#         del schema["type"]
#     for prop in schema["properties"]:
#         if "title" in schema["properties"][prop]:
#             del schema["properties"][prop]["title"]
#     return schema
#
#
# def pydantic_chain(cls):
#     return (
#         ChatPromptTemplate.from_messages([("human", "{query}")])
#         | ChatVertexAI()
#         | StrOutputParser()
#     )
#
#
# # main_chain = (
# #         RunnableParallel(
# #             message=itemgetter("message"),
# #             history=itemgetter("history"),
# #             assistant=lambda x: x['assistant'] if 'assistant' in x else "default",
# #         )
# #         | RunnablePassthrough.assign(assistant=routing_chain)
# #         | RunnablePassthrough.assign(
# #     # RunnableLambda doesn't work here because it breaks streaming.
# #     output=RunnableSwitch(
# #         "assistant",
# #         branches = {_.name: _.chain for _ in assistants},
# #         default_branch = [_.chain for _ in assistants if _.name == 'default'][0]
# #     )
# # )
# # )
#
#
# PYDANTIC_FORMAT_INSTRUCTIONS = """The output should be formatted as a JSON instance that conforms to the JSON schema below.
#
# As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
# the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.
#
# Here is the output schema:
# ```
# {schema}
# ```"""
#
#
# class PydanticV2OutputParser(BaseOutputParser):
#     """Parse an output using a pydantic model."""
#
#     pydantic_object: Type[T]
#     """The pydantic model to parse."""
#
#     def parse(self, text: str) -> T:
#         try:
#             # Greedy search for 1st json candidate.
#             match = re.search(
#                 r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL
#             )
#             json_str = match.group() if match else ""
#             json_object = json.loads(json_str, strict=False)
#             return self.pydantic_object.parse_obj(json_object)
#
#         except (json.JSONDecodeError, ValidationError) as e:
#             name = self.pydantic_object.__name__
#             msg = f"Failed to parse {name} from completion {text}. Got: {e}"
#             raise OutputParserException(msg, llm_output=text)
#
#     def get_format_instructions(self) -> str:
#         schema = self.pydantic_object.schema()
#
#         # Remove extraneous fields.
#         reduced_schema = schema
#         if "title" in reduced_schema:
#             del reduced_schema["title"]
#         if "type" in reduced_schema:
#             del reduced_schema["type"]
#         # Ensure json in context is well-formed with double quotes.
#         schema_str = json.dumps(reduced_schema)
#
#         return PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)
#
#     @property
#     def _type(self) -> str:
#         return "pydantic"
#
#
# NEXT_QUESTION_PROMPT = """
# You are trying to create an "ActionItem", which is "A task that to be captured in the to-do list.".
#
# Here is the schema of the structure you need to fill in:
# {
#         "properties": {
#             "name": {"description": "A short name for the task that will be displayed in lists.", "type": "string"},
#             "description": {"description": "A detailed description of the task.", "type": "string"},
#             "assignee": {"description": "Email address of the person to whom this task is assigned.", "type": "string"},
#             "due_date": {"description": "Date on which the task is due.", "format":"date", "type": "string"},
#             "done": {"description": "Whether the task is done or not.", "type":"boolean", "default": False},
#         },
#         "required": ["name", "description", "assignee", "due_date"],
#     }
# }
#
# At the start of each message, I'll tell you what the information you already have is.
#
# Ask the user questions, one at a time, until you have all the information.
# """
#
# PARSING_PROMPT = """
# You are trying to create an "ActionItem", which is "A task that to be captured in the to-do list.".
#
# The output should be formatted as a JSON instance that conforms to the JSON schema below.
#
# As an example, for the schema {"properties": {"foo": {"title": "Foo", "description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}
# the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar", "baz"]}} is not well-formatted.
#
# Here is the output schema:
# {
#         "properties": {
#             "name": {"description": "A short name for the task that will be displayed in lists.", "type": "string"},
#             "description": {"description": "A detailed description of the task.", "type": "string"},
#             "assignee": {"description": "Email address of the person to whom this task is assigned.", "type": "string"", "format":"email"},
#             "due_date": {"description": "Date on which the task is due.", "format":"date", "type": "string"},
#             "done": {"description": "Whether the task is done or not.", "type":"boolean"},
#         },
#     }
# }
#
# Today is Wednesday, 1/10/2024
#
# Extract whatever structured information you can from the user's message.  Be conservative, if you aren't sure, just ignore the field.  Don't make up anything.  Don't include any fields in the output that you can explicitly populate based on the input.
# """



import asyncio
import logging
from typing import AsyncIterator

import asyncstdlib

from voice_stream.types import EndOfStreamMarker

logger = logging.getLogger(__name__)


class QueueWithException(asyncio.Queue):
    """
    A specialized asyncio.Queue that propagates exceptions to the consumers.

    This queue enhances the standard asyncio.Queue by handling exceptions. When an exception
    is set in the queue via `set_exception()`, it is propagated to the consumers either when they call `get()` or
    `get_nowait()`.

    Examples
    --------
    >>> queue = queue_source()
    >>> queue.put(1)
    >>> queue.put(2)
    >>> queue.set_exception(RuntimeError("Test Error"))
    >>> await array_sink(queue)
    Caught exception: Test Error

    Notes
    -----
    - If an exception is set, it maintains its place in the queue.  If there are existing items in the queue, those will
      be returned in calls to `get()` until all items in the queue when `set_exception` was called have been exhausted.
    """

    def __init__(self):
        super().__init__()
        self.exception = None

    async def enqueue_iterator(self, async_iter: AsyncIterator):
        """Enqueues items from an asynchronous iterator to the queue.

        Parameters:
        -----------
            async_iter : AsyncIterator
                An asynchronous iterator whose items will be enqueued.

        """
        if not hasattr(async_iter, "__aiter__"):
            raise ValueError(f"Object {async_iter} is not an async iterator")
        try:
            async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
                async for item in owned_aiter:
                    await self.put(item)
        except Exception as e:
            self.exception = e
        # Signal end of iteration
        await self.put(EndOfStreamMarker)

    async def get(self):
        out = await super().get()
        return self._handle_get(out)

    def get_nowait(self):
        out = super().get_nowait()
        return self._handle_get(out)

    def _handle_get(self, out):
        if out == EndOfStreamMarker and self.exception:
            exception = self.exception
            self.exception = None
            raise exception
        return out

    def set_exception(self, exception: Exception):
        """Sets an exception to be propagated to consumers.

        When an exception is set, the EndOfStreamMarker is enqueued and when that marker is removed from the queue,
        the exception will be raised.
        """
        self.exception = exception



"""
The `LangChain <https://www.langchain.com/>`_ is a framework for building LLM based applications.  VoiceStream allows
any LangChain runnable to be used to power a voice application.
"""
from __future__ import annotations

from typing import AsyncIterator, Callable, Any, Optional

import asyncstdlib
from langchain_core.runnables import Runnable

from voice_stream.core import (
    fork_step,
    Output,
)
from voice_stream.speech_to_text import (
    filter_spurious_speech_start_events_step,
    SpeechStep,
)
from voice_stream.substreams import cancelable_substream_step
from voice_stream.text_to_speech import (
    TextToSpeechStep,
    tts_with_buffer_and_rate_limit_step,
)
from voice_stream.types import Input, SourceConvertable, to_source

GenericStepFunc = Callable[[AsyncIterator[Any]], AsyncIterator[Any]]


class LangchainVoiceFlow:
    """Experimental.  Not sure we should be using this."""

    def __init__(
        self,
        text_input: AsyncIterator[str],
        text_output: AsyncIterator[str],
        audio_output: AsyncIterator[bytes],
    ):
        self.text_input = text_input
        self.text_output = text_output
        self.audio_output = audio_output

    @classmethod
    def create(
        cls,
        audio_input: AsyncIterator[bytes],
        speech_step: SpeechStep,
        chain: Runnable,
        tts_step: TextToSpeechStep,
        langchain_postprocess: Optional[GenericStepFunc] = None,
    ) -> LangchainVoiceFlow:
        stream, speech_start = speech_step(audio_input)
        speech_start = filter_spurious_speech_start_events_step(speech_start)
        stream, text_input = fork_step(stream)

        def create_output_chain(
            stream: AsyncIterator[str],
        ) -> (AsyncIterator[bytes], AsyncIterator[str]):
            stream = langchain_step(stream, chain, on_completion="")
            if langchain_postprocess:
                stream = langchain_postprocess(stream)
            # stream = log_step(stream, "Token")
            return tts_with_buffer_and_rate_limit_step(stream, tts_step)

        stream, text_output = cancelable_substream_step(
            stream, speech_start, create_output_chain
        )
        return cls(
            text_input=text_input,
            text_output=text_output,
            audio_output=stream,
        )


async def langchain_step(
    async_iter: AsyncIterator[Input],
    chain: Runnable[Input, Output],
    input_key: Optional[str] = None,
    config_key: Optional[str] = None,
    on_completion: SourceConvertable = None,
) -> AsyncIterator[Output]:
    """
    Data flow step that passes each input item to a LangChain runnable and streams the output.

    This step is used to call LLMs or run any other LangChain runnable.  It receives text
    items, processes them through a specified 'chain', and yields the resulting output
    tokens asynchronously.  If `input_key` is not specified, the input will be directly passed to the runnable.
    If `input_key` is specified, then the items coming from the source iterator must be dictionaries, and the `input_key`
    specifies which item should be passed as input to the runnable.  The `config_key` can also be used to pass configuration
    to the Runnable.

    Parameters
    ----------
    async_iter : AsyncIterator[str]
        An asynchronous iterator that provides input text items.
    chain : Runnable[Input, Output]
        A Langchain Runnable that processes each input item.
    input_key : Optional[str], optional
        If specified, the item from the incoming dictionary that should be used as the input to the Runnable.
        If empty, then the incoming item will be passed directly.
    config_key : Optional[str], optional
        If specified, the item from the incoming dictionary that should be used as the config argument to the Runnable.
        Can only be specified if input_key is specified.
    on_completion : SourceConvertable, optional
        An optional source to be converted and iterated upon completion of processing each text item.  Can be used to
        signal the end of output if that isn't clear form the output of the chain itself.

    Yields
    ------
    AsyncIterator[Output]
        An asynchronous iterator yielding the output from the LangChain Runnable.

    """
    # Note on cancelling - https://github.com/langchain-ai/langchain/issues/11959
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for text in owned_aiter:
            input = text[input_key] if input_key else text
            config = text[config_key] if config_key else None
            async for token in chain.astream(input, config=config):
                yield token
            source = to_source(on_completion)
            async with asyncstdlib.scoped_iter(source) as owned_source:
                async for item in owned_source:
                    yield item
